from __future__ import absolute_import, unicode_literals
from json import loads
from pytz import timezone
from datetime import timedelta
from timezone_field import TimeZoneField
from celery import states
from celery import schedules
from django.db import models
from django.db.models import signals
from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django_celery_beat.utils import make_aware, now
from django_celery_beat.clockedschedule import clocked
from django_celery_beat.tzcrontab import TzAwareCrontab
from django_celery_beat.validators import (
    minute_validator,
    hour_validator,
    day_of_month_validator,
    month_of_year_validator,
    day_of_week_validator
)
from django_opstasks import managers



# Create your validators here.
def validate_tasks_args(value):
    # 位置参数可读性比关键字参数差
    if value:
        raise ValidationError('请使用关键字参数')

def validate_tasks_kwargs(value):
    try:
        loads(value)
    except:
        raise ValidationError('Json格式错误, 请纠正')

DAYS = 'days'
HOURS = 'hours'
MINUTES = 'minutes'
SECONDS = 'seconds'
MICROSECONDS = 'microseconds'
PERIOD_CHOICES = (
    (DAYS, u'天'),
    (HOURS, u'小时'),
    (MINUTES, u'分钟'),
    (SECONDS, u'秒'),
    (MICROSECONDS, u'毫秒'),
)
SOLAR_SCHEDULES = [(x, x) for x in sorted(schedules.solar._all_events)]
ALL_STATES = sorted(states.ALL_STATES)
TASK_STATE_CHOICES = sorted(zip(ALL_STATES, ALL_STATES))


# Create your models here.
class Task(models.Model):
    name = models.CharField(
        primary_key=True, max_length=50, verbose_name=u'NAME',
        help_text=u'Name of the task.')
    Request = models.CharField(
        max_length=50, verbose_name=u'REQUEST', default=u'celery.worker.request:Request', null=True,
        help_text=u'Request class used, or the qualified name of one.')
    Strategy = models.CharField(
        max_length=50, verbose_name=u'STRATEGY',
        default=u'celery.worker.strategy:default', null=True,
        help_text=u'Execution strategy used, or the qualified name of one.')
    abstract = models.BooleanField(
        verbose_name=u'ABSTRACT', default=True, null=True,
        help_text=u'Deprecated attribute abstract here for compatibility.')
    acks_late = models.BooleanField(
        verbose_name=u'ACKS LATE', default=False, null=True,
        help_text=(
            u'When enabled messages for this task will be acknowledged '
            u'after the task has been executed, and not just before (the default behavior).'))
    acks_on_failure_or_timeout = models.BooleanField(
        verbose_name=u'ACKS FAILURE', default=True, null=True,
        help_text=(
            u'When enabled messages for this task will be acknowledged '
            u'even if it fails or times out.'))
    app_main = models.CharField(
        max_length=20, verbose_name=u'APP NAME', default='opstasks',
        null=True, help_text=u'Celery application.')
    autoregister = models.BooleanField(
        verbose_name=u'AUTO REGISTER', default=True, null=True,
        help_text=u'If disabled this task won’t be registered automatically.')
    backend_url = models.CharField(
        max_length=100, verbose_name=u'BACKEND URL', null=True,
        default='mysql://user:passwd@localhost/opstasks',
        help_text=u'The result store backend class, or the name of the backend class to use.')
    default_retry_delay = models.SmallIntegerField(
        verbose_name=u'DEFAULT RETRY DELAY', default=180, null=True,
        help_text=(
            u'Default time in seconds before a retry of the task should be executed. '
            u'3 minutes by default.'))
    # expires
    # from_config
    ignore_result = models.BooleanField(
        verbose_name=u'IGNORE RESULT', default=False, null=True,
        help_text=u'If enabled the worker won’t store task state and return values for this task.')
    max_retries = models.SmallIntegerField(
        verbose_name=u'MAX RETRIES', default=3, null=True,
        help_text=(
            u'Maximum number of retries before giving up. '
            u'If set to None, it will never stop retrying.'))
    priority = models.SmallIntegerField(
        verbose_name=u'PRIORITY', null=True,
        help_text='The task priority, a number between 0 and 9. .')
    rate_limit = models.CharField(
        max_length=10, verbose_name=u'RATE LIMIT', null=True,
        help_text=(
            u'None (no rate limit), "100/s" (hundred tasks a second),'
            u'"100/m" (hundred tasks a minute), "100/h" (hundred tasks an hour)'))
    reject_on_worker_lost = models.BooleanField(
        verbose_name=u'REJECT ON LOST', null=True,
        help_text=(
            u'Even if acks_late is enabled, the worker will acknowledge tasks '
            u'when the worker process executing them abruptly exits or is signaled '
            u'(e.g., KILL/INT, etc).\n'
            u'Setting this to true allows the message to be re-queued instead, '
            u'so that the task will execute again by the same worker, or another worker.\n'
            u'Warning: Enabling this can cause message loops; '
            u'make sure you know what you’re doing.'))
    # request
    resultrepr_maxsize = models.SmallIntegerField(
        verbose_name=u'RESULTREPR MAXSIZE', default=1024, null=True,
        help_text=u'Max length of result representation used in logs and events.')
    send_events = models.BooleanField(
        verbose_name=u'SEND EVENTS', default=False, null=True,
        help_text=(
            u'If enabled the worker will send monitoring events related to this task\n'
            u'but only if the worker is configured to send task related events'))
    serializer = models.CharField(
        max_length=10, verbose_name=u'SERIALIZER', default='json', null=True,
        help_text=u'The name of a serializer that are registered with kombu.serialization.registry')
    soft_time_limit = models.SmallIntegerField(
        verbose_name=u'SOFT TIME LIMIT', null=True,
        help_text=u'Soft time limit. Defaults to the task_soft_time_limit setting.')
    store_errors_even_if_ignored = models.BooleanField(
        verbose_name=u'STORE ERRORS', default=False, null=True,
        help_text=(
            u'When enabled errors will be stored '
            u'even if the task is otherwise configured to ignore results.'))
    throws = models.CharField(
        max_length=100, verbose_name=u'THROWS', default='()', null=True,
        help_text=u'Tuple of expected exceptions.')
    time_limit = models.SmallIntegerField(
        verbose_name=u'TIME LIMIT', null=True,
        help_text=u'Hard time limit. Defaults to the task_time_limit setting.')
    track_started = models.BooleanField(
        verbose_name=u'TRACK STARTED', default=False, null=True,
        help_text=(
            u'If enabled the task will report its status as ‘started’ '
            u'when the task is executed by a worker.'))
    trail = models.BooleanField(
        verbose_name=u'TRAIL', default=True, null=True,
        help_text=(
            u'If enabled the request will keep track of subtasks started by this task, '
            u'and this information will be sent with the result (result.children).'))
    typing = models.BooleanField(
        verbose_name=u'TYPING', default=True, null=True,
        help_text=(
            u'Enable argument checking. '
            u'You can set this to false if you don’t want the signature to be checked '
            u'when calling the task. Defaults to Celery.strict_typing.'))
    description = models.CharField(
        max_length=255, verbose_name=u'DESCRIPTION', unique=False, null=True)
    timestamp = models.DateTimeField(
        verbose_name=u'TIMESTAMP', auto_now=True,
        help_text=u'Last update time of the task')
    status_choices = ((0, u'NORMAL'), (1, u'INVALID'))
    status = models.IntegerField(
        verbose_name=u'STATUS', null=False, unique=False, default=0, choices=status_choices)

    class Meta:
        verbose_name = u'任务列表'
        verbose_name_plural = verbose_name
        db_table = "tasks_task"

    def __str__(self):
        return self.name


class SolarSchedule(models.Model):
    """Schedule following astronomical patterns.

    Example: to run every sunrise in New York City:
    event='sunrise', latitude=40.7128, longitude=74.0060
    """

    event = models.CharField(
        max_length=24, choices=SOLAR_SCHEDULES,
        verbose_name=u'Solar Event',
        help_text=u'The type of solar event when the job should run')
    latitude = models.DecimalField(
        max_digits=9, decimal_places=6, verbose_name=u'纬度',
        help_text=u'Run the task when the event happens at this latitude',
        validators=[MinValueValidator(-90), MaxValueValidator(90)],)
    longitude = models.DecimalField(
        max_digits=9, decimal_places=6, verbose_name=u'经度',
        help_text=u'Run the task when the event happens at this longitude',
        validators=[MinValueValidator(-180), MaxValueValidator(180)],)

    class Meta:
        verbose_name = u'solar'
        verbose_name_plural = u'solars'
        ordering = ('event', 'latitude', 'longitude')
        unique_together = ('event', 'latitude', 'longitude')
        db_table = "tasks_solar_schedule"

    @property
    def schedule(self):
        return schedules.solar(
            self.event, self.latitude, self.longitude, nowfun=lambda: make_aware(now()))

    @classmethod
    def from_schedule(cls, schedule):
        spec = {
            'event': schedule.event,
            'latitude': schedule.lat,
            'longitude': schedule.lon}
        try:
            return cls.objects.get(**spec)
        except cls.DoesNotExist:
            return cls(**spec)
        except MultipleObjectsReturned:
            cls.objects.filter(**spec).delete()
            return cls(**spec)

    def __str__(self):
        return '{0} ({1}, {2})'.format(
            self.get_event_display(),
            self.latitude,
            self.longitude
        )


class IntervalSchedule(models.Model):
    """
    Schedule executing on a regular interval.

    Example: execute every 2 days
    every=2, period=DAYS
    """

    DAYS = DAYS
    HOURS = HOURS
    MINUTES = MINUTES
    SECONDS = SECONDS
    MICROSECONDS = MICROSECONDS

    PERIOD_CHOICES = PERIOD_CHOICES

    every = models.IntegerField(
        null=False, verbose_name=u'间隔时间', validators=[MinValueValidator(1)],
        help_text=u'再次运行任务之前要等待的间隔时间')
    period = models.CharField(
        max_length=24, choices=PERIOD_CHOICES, verbose_name=u'时间单位')

    class Meta:
        verbose_name = u'interval'
        verbose_name_plural = u'intervals'
        ordering = ['period', 'every']
        unique_together = (('period', 'every'),)
        db_table = "tasks_interval_schedule"

    @property
    def schedule(self):
        return schedules.schedule(
            timedelta(**{self.period: self.every}),
            nowfun=lambda: make_aware(now()))

    @classmethod
    def from_schedule(cls, schedule, period=SECONDS):
        every = max(schedule.run_every.total_seconds(), 0)
        try:
            return cls.objects.get(every=every, period=period)
        except cls.DoesNotExist:
            return cls(every=every, period=period)
        except MultipleObjectsReturned:
            cls.objects.filter(every=every, period=period).delete()
            return cls(every=every, period=period)

    def __str__(self):
        readable_period = None
        if self.every == 1:
            for period, _readable_period in PERIOD_CHOICES:
                if period == self.period:
                    readable_period = _readable_period.lower()
                    break
            return u'每 {} 执行一次'.format(readable_period)
        for period, _readable_period in PERIOD_CHOICES:
            if period == self.period:
                readable_period = _readable_period.lower()
                break
        return u'每 {} {} 执行一次'.format(self.every, readable_period)

    @property
    def period_singular(self):
        return self.period[:-1]


class ClockedSchedule(models.Model):
    """
    clocked schedule.
    """

    clocked_time = models.DateTimeField(
        verbose_name=u'Clock Time', help_text=u'在指定时间运行任务.')
    enabled = models.BooleanField(
        default=True, editable=False, verbose_name='ENABLED',
        help_text=u'设置为False后不进行任务调度.')

    class Meta:
        verbose_name = u'clocked'
        verbose_name_plural = u'clocked'
        ordering = ['clocked_time']
        db_table = "tasks_clocked_schedule"

    def __str__(self):
        _local_tz = timezone(settings.TIME_ZONE)
        return f'{self.clocked_time.astimezone(_local_tz)}'

    @property
    def schedule(self):
        _clocked = clocked(
            clocked_time=self.clocked_time, enabled=self.enabled, model=self)
        return _clocked

    @classmethod
    def from_schedule(cls, schedule):
        spec = {
            'clocked_time': schedule.clocked_time,
            'enabled': schedule.enabled,}
        try:
            return cls.objects.get(**spec)
        except cls.DoesNotExist:
            return cls(**spec)
        except MultipleObjectsReturned:
            cls.objects.filter(**spec).delete()
            return cls(**spec)


class CrontabSchedule(models.Model):
    """
    Timezone Aware Crontab-like schedule.

    Example:  Run every hour at 0 minutes for days of month 10-15
    minute="0", hour="*", day_of_week="*",
    day_of_month="10-15", month_of_year="*"
    """

    # The worst case scenario for day of month is a list of all 31 day numbers
    # '[1, 2, ..., 31]' which has a length of 115. Likewise, minute can be
    # 0..59 and hour can be 0..23. Ensure we can accomodate these by allowing
    # 4 chars for each value (what we save on 0-9 accomodates the []).
    # We leave the other fields at their historical length.

    minute = models.CharField(
        max_length=60 * 4, default='*', verbose_name=u'Minutes',
        validators=[minute_validator],)
    hour = models.CharField(
        max_length=24 * 4, default='*', verbose_name=u'Hours',
        validators=[hour_validator],)
    day_of_week = models.CharField(
        max_length=64, default='*', verbose_name=u'Days Of The Week',
        validators=[day_of_week_validator],)
    day_of_month = models.CharField(
        max_length=31 * 4, default='*', verbose_name=u'Days Of The Month',
        validators=[day_of_month_validator],)
    month_of_year = models.CharField(
        max_length=64, default='*', verbose_name=u'Months Of The Year',
        validators=[month_of_year_validator],)
    timezone = TimeZoneField(
        default='UTC', verbose_name='Timezone')

    class Meta:
        verbose_name = u'crontab'
        verbose_name_plural = u'crontabs'
        ordering = [
            'month_of_year', 'day_of_month', 'day_of_week', 'hour', 'minute', 'timezone']
        db_table = "tasks_crontab_schedule"

    def __str__(self):
        _data = {
            "m": self.cronexp(self.minute),
            "h": self.cronexp(self.hour),
            "dm": self.cronexp(self.day_of_month),
            "my": self.cronexp(self.month_of_year),
            "dw": self.cronexp(self.day_of_week)
        }
        return 'Crontab({m} {h} {dm} {my} {dw})'.format(**_data)

    def cronexp(self, field):
        """
        Representation of cron expression.
        """
        return field and str(field).replace(' ', '') or '*'

    @property
    def schedule(self):
        crontab = schedules.crontab(
            minute=self.minute,
            hour=self.hour,
            day_of_week=self.day_of_week,
            day_of_month=self.day_of_month,
            month_of_year=self.month_of_year,
        )
        if getattr(settings, 'DJANGO_CELERY_BEAT_TZ_AWARE', True):
            crontab = TzAwareCrontab(
                minute=self.minute,
                hour=self.hour,
                day_of_week=self.day_of_week,
                day_of_month=self.day_of_month,
                month_of_year=self.month_of_year,
                tz=self.timezone
            )
        return crontab

    @classmethod
    def from_schedule(cls, schedule):
        spec = {
            'minute': schedule._orig_minute,
            'hour': schedule._orig_hour,
            'day_of_week': schedule._orig_day_of_week,
            'day_of_month': schedule._orig_day_of_month,
            'month_of_year': schedule._orig_month_of_year,
            'timezone': schedule.tz}
        try:
            return cls.objects.get(**spec)
        except cls.DoesNotExist:
            return cls(**spec)
        except MultipleObjectsReturned:
            cls.objects.filter(**spec).delete()
            return cls(**spec)


class PeriodicTasks(models.Model):
    """
    Helper table for tracking updates to periodic tasks.

    This stores a single row with ident=1.  last_update is updated
    via django signals whenever anything is changed in the PeriodicTask model.
    Basically this acts like a DB data audit trigger.
    Doing this so we also track deletions, and not just insert/update.
    """

    ident = models.SmallIntegerField(default=1, primary_key=True, unique=True)
    last_update = models.DateTimeField(null=False)

    objects = managers.ExtendedManager()

    class Meta:
        db_table = "tasks_periodic_tasks"

    @classmethod
    def changed(cls, instance, **kwargs):
        if not instance.no_changes:
            cls.update_changed()
        del kwargs

    @classmethod
    def update_changed(cls, **kwargs):
        cls.objects.update_or_create(ident=1, defaults={'last_update': now()})
        del kwargs

    @classmethod
    def last_change(cls):
        try:
            return cls.objects.get(ident=1).last_update
        except cls.DoesNotExist:
            pass


class PeriodicTask(models.Model):
    """
    Model representing a periodic task.
    """
    name = models.CharField(max_length=200, unique=True, verbose_name=u'SHORT DESCRIPTION NAME')
    task = models.CharField(max_length=200, verbose_name=u'TASK')
    registered_task = models.ForeignKey(
        Task, on_delete=models.DO_NOTHING, to_field='name', verbose_name=u'已注册任务',
        null=True, blank=True, related_name='TASK')
    interval = models.ForeignKey(
        IntervalSchedule, on_delete=models.CASCADE, null=True, blank=True,
        verbose_name=u'Interval计划', help_text=u'注:仅能设置一种调度方式,请将其余计划留空.')
    crontab = models.ForeignKey(
        CrontabSchedule, on_delete=models.CASCADE, null=True, blank=True,
        verbose_name=u'Crontab计划', help_text=u'注:仅能设置一种调度方式,请将其余计划留空.')
    solar = models.ForeignKey(
        SolarSchedule, on_delete=models.CASCADE, null=True, blank=True,
        verbose_name=u'Solar计划', help_text=u'注:仅能设置一种调度方式,请将其余计划留空.')
    clocked = models.ForeignKey(
        ClockedSchedule, on_delete=models.CASCADE, null=True, blank=True,
        verbose_name=u'Clocked计划', help_text=u'注:仅能设置一种调度方式,请将其余计划留空.')
    args = models.TextField(
        blank=True, verbose_name=u'位置参数', validators=[validate_tasks_args],
        help_text=u'传递给任务的位置参数, Json格式')
    kwargs = models.TextField(
        blank=True, verbose_name=u'任务参数', validators=[validate_tasks_kwargs],
        help_text=u'传递给任务的关键字参数, Json格式')
    queue = models.CharField(
        max_length=200, blank=True, null=True, default=None, verbose_name=u'队列',
        help_text=u'选择CELERY_TASK_QUEUES中定义的队列, 留空将使用默认队列.')
    # you can use low-level AMQP routing options here,
    # but you almost certaily want to leave these as None
    # http://docs.celeryproject.org/en/latest/userguide/routing.html#exchanges-queues-and-routing-keys
    exchange = models.CharField(
        max_length=200, blank=True, null=True, default=None,
        verbose_name=u'Exchange', help_text=u'Override Exchange for low-level AMQP routing')
    routing_key = models.CharField(
        max_length=200, blank=True, null=True, default=None, verbose_name=u'Routing Key',
        help_text=u'Override Routing Key for low-level AMQP routing')
    headers = models.TextField(
        blank=True, default='{}', verbose_name=u'AMQP Message Headers',
        help_text=u'JSON encoded message headers for the AMQP message.')
    priority = models.PositiveIntegerField(
        default=None, validators=[MaxValueValidator(255)], blank=True, null=True,
        verbose_name=u'优先级',
        help_text=u'优先级数字介于0-255之间. 仅支持: RabbitMQ, Redis . 优先级相反, 0最高.')
    start_time = models.DateTimeField(
        blank=True, null=True, verbose_name=u'开始时间', help_text=u'计划的生效时间')
    expires = models.DateTimeField(
        blank=True, null=True, verbose_name='过期时间',
        help_text=u'超过该时间之后, 本计划将不再启动任务')
    expire_seconds = models.PositiveIntegerField(
        blank=True, null=True, verbose_name=u'Expires timedelta with seconds',
        help_text=(
            u'Timedelta with seconds which the schedule will no longer '
            u'trigger the task to run'))
    enabled = models.BooleanField(
        default=True, verbose_name=u'ENABLED', help_text=u'取消勾选将关闭计划任务')
    one_off = models.BooleanField(
        default=False, verbose_name=u'ONE_OFF', help_text=u'勾选后该计划仅执行一次')
    last_run_at = models.DateTimeField(
        auto_now=False, auto_now_add=False, editable=False, blank=True, null=True,
        verbose_name=u'Last Run Datetime')
    total_run_count = models.PositiveIntegerField(
        default=0, editable=False, verbose_name=u'Total Run Count',
        help_text=(
            u'Running count of how many times the schedule '
            u'has triggered the task'))
    date_changed = models.DateTimeField(
        auto_now=True, verbose_name=u'Last Modified',
        help_text=u'Datetime that this PeriodicTask was last modified')
    description = models.TextField(
        blank=True, verbose_name=u'DESCRIPTION',
        default=u'Detailed description about the details of this Periodic Task')
    objects = managers.PeriodicTaskManager()
    no_changes = False

    class Meta:
        verbose_name = u'定时任务'
        verbose_name_plural = verbose_name
        db_table = "tasks_periodic_task"

    def validate_unique(self, exclude=None):
        """
        Check unique constraints on the model and raise ValidationError if any failed.
        """
        super(PeriodicTask, self).validate_unique(exclude=exclude)
        schedule_types = ['interval', 'crontab', 'solar', 'clocked']
        selected_schedule_types = [
            s for s in schedule_types if getattr(self, s)]
        if len(selected_schedule_types) == 0:
            raise ValidationError({
                'interval': ['One of clocked, interval, crontab, or solar must be set.']})
        err_msg = 'Only one of clocked, interval, crontab, or solar must be set'
        if len(selected_schedule_types) > 1:
            error_info = {}
            for selected_schedule_type in selected_schedule_types:
                error_info[selected_schedule_type] = [err_msg]
            raise ValidationError(error_info)

        # clocked must be one off task
        if self.clocked and not self.one_off:
            err_msg = 'clocked must be one off, one_off must set True'
            raise ValidationError(err_msg)

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        """
        Save the current instance. Override this in a subclass if you want to
        control the saving process.

        The 'force_insert' and 'force_update' parameters can be used to insist
        that the "save" must be an SQL insert or update (or equivalent for
        non-SQL backends), respectively. Normally, they should not be set.
        """
        self.exchange = self.exchange or None
        self.routing_key = self.routing_key or None
        self.queue = self.queue or None
        self.headers = self.headers or None
        if not self.enabled:
            self.last_run_at = None
        self._clean_expires()
        self.validate_unique()
        super(PeriodicTask, self).save(
            force_insert=force_insert, force_update=force_update,
            using=using, update_fields=update_fields)

    def _clean_expires(self):
        if self.expire_seconds is not None and self.expires:
            raise ValidationError('Only one can be set, in expires and expire_seconds')

    @property
    def expires_(self):
        return self.expires or self.expire_seconds

    def __str__(self):
        fmt = '{0.name}: {{no schedule}}'
        if self.interval:
            fmt = '{0.name}: {0.interval}'
        if self.crontab:
            fmt = '{0.name}: {0.crontab}'
        if self.solar:
            fmt = '{0.name}: {0.solar}'
        if self.clocked:
            fmt = '{0.name}: {0.clocked}'
        return fmt.format(self)

    @property
    def schedule(self):
        if self.interval:
            return self.interval.schedule
        if self.crontab:
            return self.crontab.schedule
        if self.solar:
            return self.solar.schedule
        if self.clocked:
            return self.clocked.schedule


class TaskResult(models.Model):
    """
    Task result/status.
    """
    task_id = models.CharField(
        max_length=getattr(settings, 'DJANGO_CELERY_RESULTS_TASK_ID_MAX_LENGTH', 255),
        unique=True, db_index=True, verbose_name=u'Task ID')
    task_name = models.CharField(
        null=True, max_length=255, db_index=True, verbose_name=u'Task Name')
    task_args = models.TextField(
        null=True, verbose_name=u'Task Positional Arguments')
    task_kwargs = models.TextField(
        null=True, verbose_name=u'Task Named Arguments')
    status = models.CharField(
        max_length=50, default=states.PENDING, db_index=True,
        choices=TASK_STATE_CHOICES, verbose_name=u'Task State')
    worker = models.CharField(
        max_length=100, db_index=True, default=None, null=True, verbose_name=u'Worker')
    content_type = models.CharField(
        blank=True, null=True, max_length=128, verbose_name=u'Result Content Type')
    content_encoding = models.CharField(
        blank=True, null=True, max_length=64, verbose_name=u'Result Encoding')
    result = models.TextField(
        blank=True, null=True, default=None, editable=False, verbose_name=u'Result Data')
    date_start = models.DateTimeField(
        auto_now_add=True, blank=True, null=True, verbose_name=u'Started DateTime')
    date_created = models.DateTimeField(
        auto_now_add=True, blank=True, null=True, db_index=True,
        verbose_name=u'Created DateTime')
    date_done = models.DateTimeField(
        auto_now=True, blank=True, null=True, db_index=True,
        verbose_name=u'Completed DateTime')
    traceback = models.TextField(
        blank=True, null=True, verbose_name=u'Traceback')
    meta = models.TextField(
        null=True, default=None, editable=False, verbose_name=u'Task Meta Information')
    queue = models.CharField(max_length=155, blank=True, null=True, verbose_name=u'Queue')

    objects = managers.TaskResultManager()

    class Meta:
        verbose_name = u'执行记录'
        verbose_name_plural = verbose_name
        db_table = "tasks_results"


    def as_dict(self):
        return {
            'task_id': self.task_id,
            'task_name': self.task_name,
            'task_args': self.task_args,
            'task_kwargs': self.task_kwargs,
            'status': self.status,
            'result': self.result,
            'date_done': self.date_done,
            'traceback': self.traceback,
            'meta': self.meta,
            'worker': self.worker
        }

    def __str__(self):
        return '<Task: {0.task_id} ({0.status})>'.format(self)


class TaskStartDatetime(models.Model):
    task_id = models.CharField(
        max_length=getattr(settings, 'DJANGO_CELERY_RESULTS_TASK_ID_MAX_LENGTH', 255),
        unique=True, db_index=True, verbose_name=u'Task ID')
    date_start = models.DateTimeField(
        auto_now_add=True, blank=True, null=True, verbose_name=u'Started DateTime')

    class Meta:
        db_table = "tasks_start_datetime"


signals.pre_delete.connect(PeriodicTasks.changed, sender=PeriodicTask)
signals.pre_save.connect(PeriodicTasks.changed, sender=PeriodicTask)
signals.pre_delete.connect(
    PeriodicTasks.update_changed, sender=IntervalSchedule)
signals.post_save.connect(
    PeriodicTasks.update_changed, sender=IntervalSchedule)
signals.post_delete.connect(
    PeriodicTasks.update_changed, sender=CrontabSchedule)
signals.post_save.connect(
    PeriodicTasks.update_changed, sender=CrontabSchedule)
signals.post_delete.connect(
    PeriodicTasks.update_changed, sender=SolarSchedule)
signals.post_save.connect(
    PeriodicTasks.update_changed, sender=SolarSchedule)
signals.post_delete.connect(
    PeriodicTasks.update_changed, sender=ClockedSchedule)
signals.post_save.connect(
    PeriodicTasks.update_changed, sender=ClockedSchedule)



from celery import Task
from logging import getLogger
from django_opstasks.models import TaskResult

LOGGER = getLogger('django')


def record_start_datetime(func):
    """
    Decorator: record task start time
    """
    def _inner(*args, **kwargs):
        task_name = args[0].name
        LOGGER.info('start %s', task_name)
        _result = func(*args, **kwargs)
        _data = {
            "task_id": _result.task_id,
        }
        TaskResult.objects.update_or_create(**_data)
        return _result
    return _inner


class CustomTask(Task):
    """ Custom Base Task
    """

    def run(self, *args, **kwargs):
        """The body of the task executed by workers."""
        raise NotImplementedError('Tasks must define the run method.')

    @record_start_datetime
    def apply(self, args=None, kwargs=None,
              link=None, link_error=None, task_id=None, retries=None,
              throw=None, logfile=None, loglevel=None, headers=None, **options):
        """
        Execute this task locally, by blocking until the task returns.

          Arguments:
            args (Tuple): positional arguments passed on to the task.
            kwargs (Dict): keyword arguments passed on to the task.
            throw (bool): Re-raise task exceptions.
                Defaults to the :setting:`task_eager_propagates` setting.

          Returns:
            celery.result.EagerResult: pre-evaluated result.
        """
        LOGGER.warning("task should not be started in apply, forward to apply_async")
        return super().apply_async(
            args=args, kwargs=kwargs, task_id=task_id,
            link=link, link_error=link_error, **options)

    @record_start_datetime
    def apply_async(self, args=None, kwargs=None, task_id=None,
                    producer=None, link=None, link_error=None, shadow=None, **options):
        """Apply tasks asynchronously by sending a message.

        Arguments:
            args (Tuple): The positional arguments to pass on to the task.

            kwargs (Dict): The keyword arguments to pass on to the task.

            countdown (float): Number of seconds into the future that the
                task should execute.  Defaults to immediate execution.

            eta (~datetime.datetime): Absolute time and date of when the task
                should be executed.  May not be specified if `countdown`
                is also supplied.

            expires (float, ~datetime.datetime): Datetime or
                seconds in the future for the task should expire.
                The task won't be executed after the expiration time.

            shadow (str): Override task name used in logs/monitoring.
                Default is retrieved from :meth:`shadow_name`.

            connection (kombu.Connection): Re-use existing broker connection
                instead of acquiring one from the connection pool.

            retry (bool): If enabled sending of the task message will be
                retried in the event of connection loss or failure.
                Default is taken from the :setting:`task_publish_retry`
                setting.  Note that you need to handle the
                producer/connection manually for this to work.

            retry_policy (Mapping): Override the retry policy used.
                See the :setting:`task_publish_retry_policy` setting.

            queue (str, kombu.Queue): The queue to route the task to.
                This must be a key present in :setting:`task_queues`, or
                :setting:`task_create_missing_queues` must be
                enabled.  See :ref:`guide-routing` for more
                information.

            exchange (str, kombu.Exchange): Named custom exchange to send the
                task to.  Usually not used in combination with the ``queue``
                argument.

            routing_key (str): Custom routing key used to route the task to a
                worker server.  If in combination with a ``queue`` argument
                only used to specify custom routing keys to topic exchanges.

            priority (int): The task priority, a number between 0 and 9.
                Defaults to the :attr:`priority` attribute.

            serializer (str): Serialization method to use.
                Can be `pickle`, `json`, `yaml`, `msgpack` or any custom
                serialization method that's been registered
                with :mod:`kombu.serialization.registry`.
                Defaults to the :attr:`serializer` attribute.

            compression (str): Optional compression method
                to use.  Can be one of ``zlib``, ``bzip2``,
                or any custom compression methods registered with
                :func:`kombu.compression.register`.
                Defaults to the :setting:`task_compression` setting.

            link (Signature): A single, or a list of tasks signatures
                to apply if the task returns successfully.

            link_error (Signature): A single, or a list of task signatures
                to apply if an error occurs while executing the task.

            producer (kombu.Producer): custom producer to use when publishing
                the task.

            add_to_parent (bool): If set to True (default) and the task
                is applied while executing another task, then the result
                will be appended to the parent tasks ``request.children``
                attribute.  Trailing can also be disabled by default using the
                :attr:`trail` attribute

            publisher (kombu.Producer): Deprecated alias to ``producer``.

            headers (Dict): Message headers to be included in the message.

        Returns:
            celery.result.AsyncResult: Promise of future evaluation.

        Raises:
            TypeError: If not enough arguments are passed, or too many
                arguments are passed.  Note that signature checks may
                be disabled by specifying ``@task(typing=False)``.
            kombu.exceptions.OperationalError: If a connection to the
               transport cannot be made, or if the connection is lost.

        Note:
            Also supports all keyword arguments supported by
            :meth:`kombu.Producer.publish`.
        """
        return super().apply_async(
            args=args, kwargs=kwargs, task_id=task_id,
            producer=producer, link=link, link_error=link_error, shadow=shadow, **options)



# Generated by Django 3.0.8 on 2020-09-01 14:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_opstasks', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskresult',
            name='date_start',
            field=models.DateTimeField(auto_now_add=True, null=True, verbose_name='Started DateTime'),
        ),
        migrations.AddField(
            model_name='taskresult',
            name='queue',
            field=models.CharField(blank=True, max_length=155, null=True, verbose_name='Queue'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='content_encoding',
            field=models.CharField(blank=True, max_length=64, null=True, verbose_name='Result Encoding'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='content_type',
            field=models.CharField(blank=True, max_length=128, null=True, verbose_name='Result Content Type'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Created DateTime'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='date_done',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Completed DateTime'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='meta',
            field=models.TextField(default=None, editable=False, null=True, verbose_name='Task Meta Information'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='result',
            field=models.TextField(blank=True, default=None, editable=False, null=True, verbose_name='Result Data'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='status',
            field=models.CharField(choices=[('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('RECEIVED', 'RECEIVED'), ('RETRY', 'RETRY'), ('REVOKED', 'REVOKED'), ('STARTED', 'STARTED'), ('SUCCESS', 'SUCCESS')], db_index=True, default='PENDING', max_length=50, verbose_name='Task State'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='task_args',
            field=models.TextField(null=True, verbose_name='Task Positional Arguments'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='task_id',
            field=models.CharField(db_index=True, max_length=255, unique=True, verbose_name='Task ID'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='task_kwargs',
            field=models.TextField(null=True, verbose_name='Task Named Arguments'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='task_name',
            field=models.CharField(db_index=True, max_length=255, null=True, verbose_name='Task Name'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='traceback',
            field=models.TextField(blank=True, null=True, verbose_name='Traceback'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='worker',
            field=models.CharField(db_index=True, default=None, max_length=100, null=True, verbose_name='Worker'),
        ),
    ]



from time import sleep
from celery import current_app
from celery.utils.log import get_task_logger
from django_opstasks.database import TasksDatabase

LOGGER = get_task_logger('django')


# 以下为测试任务
@current_app.task(bind=True, priority=1, name='opstasks.test.add')
def add(self, inta, intb):
    """测试带参数任务"""
    return int(inta) + int(intb)


@current_app.task(bind=True, priority=1, name='opstasks.test.error')
def error(self):
    """测试错误任务"""
    raise TypeError(self.__doc__)


@current_app.task(bind=True, priority=1, name='opstasks.test.longtime')
def longtime(self):
    """测试长时间任务"""
    sleep(10)
    return True


@current_app.task(bind=True, priority=8, name='opstasks.sync_task_to_database')
def sync_task_to_database(self):
    """将已注册的任务同步到数据库"""
    TasksDatabase().sync_to_database(current_app)


# Execute this task once when starting worker and beat and first importing task from opstasks.tasks
sync_task_to_database.delay()


# TODO: migrate task result start datetime



from contextlib import contextmanager
from vine.utils import wraps
from django.conf import settings
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.exc import DatabaseError, InvalidRequestError
from sqlalchemy.orm.exc import StaleDataError
from celery.utils.log import get_task_logger

LOGGER = get_task_logger('django')


def retry_when_database_error(fun):
    """
    decorator:
        retry three times when the database operation fails.
        default max_retries = 3
    """
    @wraps(fun)
    def _inner(*args, **kwargs):
        max_retries = kwargs.pop('max_retries', 3)
        for retries in range(max_retries):
            try:
                return fun(*args, **kwargs)
            except (DatabaseError, InvalidRequestError, StaleDataError):
                LOGGER.warning(
                    'failed operation %s.  retrying %s more times.',
                    fun.__name__, max_retries - retries - 1,
                    exc_info=True)
                if retries + 1 >= max_retries:
                    raise
    return _inner


class BaseDatabase(object):
    """ basic database class with context manager
    """
    def __init__(self, dburi=None):
        self.dburi = dburi if dburi else settings.DB_URI
        self.engine = None
        self.session = None

    def _create_session(self):
        try:
            self.engine = create_engine(self.dburi, echo=False, encoding='utf-8')
            self.session = sessionmaker(bind=self.engine)()
            return self.engine, self.session
        except Exception as error:
            LOGGER.error('can not create database session. {"error": %s}', error)

    def _map_table(self, table):
        '''mapping model and data table, return a model'''
        try:
            base = automap_base()
            base.prepare(self.engine, reflect=True)
            return base.classes[table]
        except Exception as error:
            LOGGER.error('failed to map model and data table. {"error": %s}', error)
            raise

    @contextmanager
    def database_connection(self):
        """ a context manager
        """
        try:
            self.engine, self.session = self._create_session()
            yield
        except Exception as error:
            self.session.rollback()
            LOGGER.error('database operation failed, has been rolled back. {"error": %s}', error)
            raise
        finally:
            self.session.close()



# Generated by Django 3.0.8 on 2020-09-01 14:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_opstasks', '0002_auto_20200901_2215'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskresult',
            name='date_created',
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True, verbose_name='Created DateTime'),
        ),
        migrations.AlterField(
            model_name='taskresult',
            name='date_done',
            field=models.DateTimeField(auto_now=True, db_index=True, null=True, verbose_name='Completed DateTime'),
        ),
    ]



"""Celery cache backend using the Django Cache Framework."""
from __future__ import absolute_import, unicode_literals

from django.core.cache import cache as default_cache, caches

from celery.backends.base import KeyValueStoreBackend


class CacheBackend(KeyValueStoreBackend):
    """Backend using the Django cache framework to store task metadata."""

    def __init__(self, *args, **kwargs):
        super(CacheBackend, self).__init__(*args, **kwargs)

        # Must make sure backend doesn't convert exceptions to dict.
        self.serializer = 'pickle'

    def get(self, key):
        return self.cache_backend.get(key)

    def set(self, key, value):
        self.cache_backend.set(key, value, self.expires)

    def delete(self, key):
        self.cache_backend.delete(key)

    def encode(self, data):
        return data

    def decode(self, payload):
        return payload

    @property
    def cache_backend(self):
        backend = self.app.conf.cache_backend
        return caches[backend] if backend else default_cache

    def add_to_chord(self, chord_id, result):
        raise NotImplementedError('Backend does not support add_to_chord')



from datetime import datetime

def current_datetime(utc=True):
    """
    returns the current utc time
    """
    formatter = '%Y-%m-%d %H:%M:%S.%f'
    if utc:
        return datetime.utcnow().strftime(formatter)
    return datetime.now().strftime(formatter)



from logging import getLogger
from django.contrib import admin, messages
from django.utils.html import format_html
from django.conf import settings
from django.template.defaultfilters import pluralize
from import_export import resources
from import_export.admin import ImportExportModelAdmin
from django_celery_beat.utils import is_database_scheduler
from .models import Task, TaskResult
from .models import (
    SolarSchedule, IntervalSchedule, CrontabSchedule,
    ClockedSchedule, PeriodicTask, PeriodicTasks
)

LOGGER = getLogger('django')

DISPLAY_NAME_FORMAT = '<div title="{}">{}</div>'
DISPLAY_STATUS_FORMAT = '<span title="{}" \
    class="el-tag el-tag--mini el-tag--light el-tag--{}">{}</span>'

# Register your models here.
class TaskResource(resources.ModelResource):
    class Meta:
        model = Task
        skip_unchanged = True
        report_skipped = True
@admin.register(Task)
class TaskAdmin(ImportExportModelAdmin):
    model = Task
    resource_class = TaskResource
    list_per_page = 20
    list_display = (
        'dis_name_description', 'description', 'dis_average_duration', 'dis_success_ratio',
        'dis_status')
    search_fields = ('name', 'description', 'status')
    list_filter = ('name', 'max_retries', 'priority', 'status')
    empty_value_display = '---'

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    # def has_delete_permission(self, request, obj=None):
    #     return False

    def dis_average_duration(self, obj):
        duration_list = []
        for result in TaskResult.objects.filter(task_name=obj.name):
            try:
                duration = result.date_done - result.date_start
                duration_list.append(duration.seconds)
            except Exception:
                pass
        if duration_list:
            min_duration, max_duration = min(duration_list), max(duration_list)
            sum_duration = 0
            for duration in duration_list:
                sum_duration += duration
            avg_duration = round(sum_duration/len(duration_list), 2)
        else:
            min_duration, avg_duration, max_duration = 0, 0, 0
        return '%s / %s / %s' % (min_duration, avg_duration, max_duration)
    dis_average_duration.short_description = u'DURATION(MIN/AVG/MAX)'
    dis_average_duration.admin_order_field = 'id'

    def dis_success_ratio(self, obj):
        status_list = [result.status for result in TaskResult.objects.filter(task_name=obj.name)]
        successed = status_list.count('SUCCESS')
        return '%s / %s' % (successed, len(status_list))
    dis_success_ratio.short_description = u'SUCCESS/TOTAL'
    dis_success_ratio.admin_order_field = 'id'

    def dis_name_description(self, obj):
        return format_html(DISPLAY_NAME_FORMAT.format(obj.description, obj.name))
    dis_name_description.short_description = u'NAME'
    dis_name_description.admin_order_field = 'name'

    def dis_status(self, obj):
        timestamp = obj.timestamp.strftime('%Y-%m-%d %H:%M:%S') if obj.timestamp else obj.status
        if obj.status:
            return format_html(DISPLAY_STATUS_FORMAT.format(timestamp, 'danger', 'INVALID'))
        return format_html(DISPLAY_STATUS_FORMAT.format(timestamp, 'success', 'NORMAL'))
    dis_status.short_description = ' '
    dis_status.admin_order_field = 'status'


@admin.register(SolarSchedule)
class SolarScheduleAdmin(admin.ModelAdmin):

    def has_module_permission(self, request):
        return False

    class Meta:
        model = SolarSchedule


@admin.register(IntervalSchedule)
class IntervalScheduleAdmin(admin.ModelAdmin):
    fields = ('every', 'period',)

    def has_module_permission(self, request):
        return False

    class Meta:
        model = IntervalSchedule


@admin.register(CrontabSchedule)
class CrontabScheduleAdmin(admin.ModelAdmin):

    def has_module_permission(self, request):
        return False

    class Meta:
        model = CrontabSchedule


@admin.register(ClockedSchedule)
class ClockedScheduleAdmin(admin.ModelAdmin):
    """
    Admin-interface for clocked schedules.
    """
    fields = ('clocked_time', 'enabled',)
    readonly_fields = ('enabled',)
    list_display = ('clocked_time', 'enabled',)

    def has_module_permission(self, request):
        return False

    class Meta:
        model = ClockedSchedule


@admin.register(PeriodicTask)
class PeriodicTaskAdmin(admin.ModelAdmin):
    """
    Admin interface for periodic tasks.
    """
    model = PeriodicTask
    list_display = (
        'dis_name_description', 'task', 'dis_scheduler', 'dis_kwargs',
        'total_run_count', 'last_run_at', 'dis_enabled')
    search_fields = ('name',)
    list_filter = ['task', 'enabled', 'one_off']
    fieldsets = (
        (u'基础配置', {
            'fields': ('name', 'registered_task', 'enabled', 'one_off', 'description',),
            'classes': ('extrapretty',),}),
        (u'调度策略', {
            'fields': ('interval', 'crontab', 'solar', 'clocked',),
            'classes': ('extrapretty',),
            'description': '请在下列四种调度计划中选择其中一种.'}),
        (u'生效时间', {
            'fields': ('start_time', 'expires'),
            'classes': ('extrapretty',)}),
        (u'任务选项', {
            'fields': ('kwargs', 'queue', 'priority'),
            'classes': ('extrapretty', 'collapse', 'in'),}),)
    readonly_fields = ('last_run_at',)
    date_hierarchy = 'start_time'
    empty_value_display = '---'
    actions = ('enable_tasks', 'disable_tasks')

    def save_model(self, request, obj, form, change):
        if obj.registered_task.status == 0:
            obj.task = obj.registered_task.name
        if obj.task:
            super().save_model(request, obj, form, change)
        else:
            messages.set_level(request, messages.WARNING)
            messages.error(request, u'当前选择的任务无效, 请确认后重新选择')

    def dis_name_description(self, obj):
        return format_html(DISPLAY_NAME_FORMAT.format(obj.description, obj.name))
    dis_name_description.short_description = u'NAME'
    dis_name_description.admin_order_field = 'name'

    def dis_scheduler(self, obj):
        if obj.interval:
            return obj.interval
        if obj.crontab:
            return obj.crontab
        if obj.clocked:
            return obj.clocked
        if obj.solar:
            return obj.solar
        return 'no schedule'
    dis_scheduler.admin_order_field = 'name'
    dis_scheduler.short_description = u'SCHEDULER'

    def dis_kwargs(self, obj):
        if obj.kwargs and obj.kwargs != '{}':
            return obj.kwargs
        return '---'
    dis_kwargs.admin_order_field = 'kwargs'
    dis_kwargs.short_description = u'KWARGS'

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        scheduler = getattr(settings, 'CELERY_BEAT_SCHEDULER', None)
        extra_context['wrong_scheduler'] = not is_database_scheduler(scheduler)
        return super(PeriodicTaskAdmin, self).changelist_view(
            request, extra_context)

    def get_queryset(self, request):
        qs = super(PeriodicTaskAdmin, self).get_queryset(request)
        return qs.select_related('interval', 'crontab', 'solar', 'clocked')

    def _message_user_about_update(self, request, rows_updated, verb):
        """ Send message about action to user.
        `verb` should shortly describe what have changed (e.g. 'enabled').
        """
        self.message_user(request, (
            f'{rows_updated} task{pluralize(rows_updated)} '
            f'{pluralize(rows_updated, "was,were")} successfully {verb}'))

    def dis_enabled(self, obj):
        if obj.enabled:
            return format_html(DISPLAY_STATUS_FORMAT.format(
                'ENABLED', 'success', 'ENABLED'))
        return format_html(DISPLAY_STATUS_FORMAT.format(
            'DISABLED', 'danger', 'DISABLED'))
    dis_enabled.short_description = u' '
    dis_enabled.admin_order_field = 'enabled'

    def enable_tasks(self, request, queryset):
        rows_updated = queryset.update(enabled=True)
        PeriodicTasks.update_changed()
        self._message_user_about_update(request, rows_updated, 'enabled')
    enable_tasks.short_description = u'开启'
    enable_tasks.type = 'success'
    enable_tasks.icon = 'el-icon-video-play'

    def disable_tasks(self, request, queryset):
        rows_updated = queryset.update(enabled=False)
        PeriodicTasks.update_changed()
        self._message_user_about_update(request, rows_updated, 'disabled')
    disable_tasks.short_description = u'关闭'
    disable_tasks.type = 'warning'
    disable_tasks.icon = 'el-icon-video-pause'
    disable_tasks.confirm = u'确定要停止任务调度吗?'

    class Meta:
        model = PeriodicTask


class TaskResultResource(resources.ModelResource):
    class Meta:
        model = TaskResult
        skip_unchanged = True
        report_skipped = True
@admin.register(TaskResult)
class TaskResultAdmin(ImportExportModelAdmin):
    fields = (
        'task_id', 'task_name', 'queue', 'worker', 'status',
        'date_start', 'date_created', 'date_done',
        'task_args', 'task_kwargs', 'traceback', 'result', 'meta')
    list_display = (
        'id', 'task_id', 'task_name', 'dis_kwargs', 'date_start',
        'queue', 'dis_duration', 'dis_colored_status')
    list_filter = ('task_name', 'worker', 'status', 'date_done')
    resource_class = TaskResultResource
    actions_selection_counter = True
    empty_value_display = '---'
    list_per_page = 100
    date_hierarchy = 'date_done'

    class Meta:
        model = TaskResult

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def history_view(self, request, object_id, extra_context=None):
        return False

    def dis_kwargs(self, obj):
        return obj.task_kwargs
    dis_kwargs.short_description = u'KWARGS'
    dis_kwargs.admin_order_field = 'kwargs'

    def dis_duration(self, obj):
        """calculate and display task execution time"""
        try:
            duration = obj.date_done - obj.date_start
            return duration.seconds
        except Exception as error:
            LOGGER.exception(error)
            return '---'
    dis_duration.short_description = u'DURATION'
    dis_duration.admin_order_field = 'date_done'

    def dis_colored_status(self, obj):
        """display: colored status of task record"""
        status_color = {
            'PENDING': 'info',
            'RECEIVED': 'warning',
            'REJECTED': 'warning',
            'IGNORED': 'warning',
            'REVOKED': 'warning',
            'STARTED': 'primary',
            'RETRY': 'warning',
            'SUCCESS': 'success',
            'FAILURE': 'danger',
            }
        date_done = obj.date_done.strftime('%Y-%m-%d %H:%M:%S') if obj.date_done else obj.status
        try:
            return format_html(DISPLAY_STATUS_FORMAT.format(
                date_done, status_color[obj.status], obj.status))
        except Exception as error:
            LOGGER.exception(error)
            return '---'
    dis_colored_status.short_description = u'STATUS'
    dis_colored_status.admin_order_field = 'status'



from json import loads
from celery.utils.log import get_task_logger
from django_opstasks.common.response import OpstasksResponse
from django_opstasks.common.response import AsyncResultResponse

LOGGER = get_task_logger('django')


# Create your views here.

def error_test(request):
    if request.method == "GET":
        from django_opstasks.tasks import error
        result = error.apply_async()
        if result.get():
            return AsyncResultResponse(result)
    return OpstasksResponse('Method Not Allowed', 405)


def args_test(request):
    if request.method == "POST":
        from django_opstasks.tasks import add
        data = loads(request.body)
        try:
            result = add.apply_async(**data)
            return AsyncResultResponse(result)
        except Exception as error:
            LOGGER.exception(error)
            return OpstasksResponse('Bad Request', 400)
    return OpstasksResponse('Method Not Allowed', 405)


def longtime_test(request):
    if request.method == "GET":
        from django_opstasks.tasks import longtime
        result = longtime.apply_async()
        return AsyncResultResponse(result)
    return OpstasksResponse('Method Not Allowed', 405)


def sync_task_to_database(request):
    if request.method == "GET":
        from django_opstasks.tasks import sync_task_to_database as task
        result = task.apply_async()
        return AsyncResultResponse(result)
    return OpstasksResponse('Method Not Allowed', 405)



# Generated by Django 3.0.8 on 2020-09-10 07:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_opstasks', '0003_auto_20200901_2231'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskStartDatetime',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.CharField(db_index=True, max_length=255, unique=True, verbose_name='Task ID')),
                ('date_start', models.DateTimeField(auto_now_add=True, null=True, verbose_name='Started DateTime')),
            ],
            options={
                'db_table': 'tasks_start_datetime',
            },
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import setuptools
import setuptools.command.test

VERSION = '0.2.3'

NAME = 'django_opstasks'

PACKAGES = setuptools.find_packages(include=(
    'django_opstasks',
    'django_opstasks.common',
    'django_opstasks.backends',
    'django_opstasks.migrations'
))

CLASSES = """
    Development Status :: 1 - Planning
    License :: OSI Approved :: BSD License
    Programming Language :: Python
    Programming Language :: Python :: 3.7
    Programming Language :: Python :: 3.8
    Programming Language :: Python :: Implementation :: CPython
    Programming Language :: Python :: Implementation :: PyPy
    Framework :: Django
    Framework :: Django :: 3.0
    Operating System :: OS Independent
    Topic :: Communications
    Topic :: System :: Distributed Computing
    Topic :: Software Development :: Libraries :: Python Modules
"""
classifiers = [s.strip() for s in CLASSES.split('\n') if s]


if sys.version_info < (3, 6):
    raise Exception(f'{NAME} {VERSION} requires Python 3.6 or later!')

with open("PYPI.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    version=VERSION,
    name=NAME,
    packages=PACKAGES,
    description='opstasks',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='opstasks',
    author='lunuan',
    author_email='admin@lunuan.com.cn',
    url='',
    platforms=['any'],
    license='BSD',
    classifiers=classifiers,
    install_requires='celery>=4.4,<5.0',
    # tests_require=reqs('test.txt') + reqs('test-django.txt'),
    # cmdclass={'test': pytest},
    entry_points={
        'celery.result_backends': [
            'opstasks-backends = django_opstasks.backends:DatabaseBackend',
            'opstasks-cache = django_opstasks.backends:CacheBackend',
        ],
    },
    zip_safe=False,
    include_package_data=False,
)



from logging import getLogger
from django_opstasks.common.datetime import current_datetime
from django_opstasks.common.sqlalchemy import retry_when_database_error
from django_opstasks.common.sqlalchemy import BaseDatabase

LOGGER = getLogger('django')


class TasksDatabase(BaseDatabase):
    """ database class based on BaseDatabase with tasks related operations
    """
    def _make_task_dict(self, task):
        """
        return a task object, exclude 'celery.*'
        """
        data = {
            "name": task.name,
            "Request": task.Request,
            "Strategy": task.Strategy,
            "abstract": task.abstract,
            "acks_late": task.acks_late,
            "acks_on_failure_or_timeout": task.acks_on_failure_or_timeout,
            "app_main": task.app.main,
            "autoregister": task.autoregister,
            "backend_url": task.backend.url,
            "default_retry_delay": task.default_retry_delay,
            "ignore_result": task.ignore_result,
            "max_retries": task.max_retries,
            "priority": task.priority,
            "rate_limit": task.rate_limit,
            "reject_on_worker_lost": task.reject_on_worker_lost,
            "resultrepr_maxsize": task.resultrepr_maxsize,
            "send_events": task.send_events,
            "serializer": task.serializer,
            "soft_time_limit": task.soft_time_limit,
            "store_errors_even_if_ignored": task.store_errors_even_if_ignored,
            "throws": str(task.throws),
            "time_limit": task.time_limit,
            "track_started": task.track_started,
            "trail": task.trail,
            "typing": task.typing,
            "description": task.__doc__,
            "timestamp": current_datetime(),
            "status": 0
        }
        return data

    @retry_when_database_error
    def sync_to_database(self, app):
        """
        synchronize registered tasks and schedule to the database.
        """
        with self.database_connection():
            tasks_model = self._map_table('tasks_task')
            LOGGER.info(
                'begin syncing registered tasks to the database, '
                'set the status of all tasks to invalid.')
            task_names = [name for name in app.tasks if name.split('.')[0] != 'celery']
            self.session.query(tasks_model).filter().update({"status": 1})
            for name in task_names:
                task = self.session.query(tasks_model).filter(tasks_model.name == name)
                if task.all():
                    LOGGER.info('task "%s" already exists, set the status to normal', name)
                    _data = self._make_task_dict(app.tasks.get(name))
                    task.update(_data)
                else:
                    LOGGER.info('add new task "%s"', name)
                    self.session.add(tasks_model(**self._make_task_dict(app.tasks.get(name))))
            self.session.commit()
        LOGGER.info('synchronize the registered tasks to the database has been completed')



from __future__ import absolute_import, unicode_literals

from celery.backends.base import BaseDictBackend
from celery.utils.serialization import b64encode, b64decode

from ..models import TaskResult


class DatabaseBackend(BaseDictBackend):
    """The Django database backend, using models to store task state."""

    TaskModel = TaskResult

    subpolling_interval = 0.5

    def _store_result(self, task_id, result, status,
                      traceback=None, request=None, using=None):
        """Store return value and status of an executed task."""
        content_type, content_encoding, result = self.encode_content(result)
        _, _, meta = self.encode_content({
            'children': self.current_task_children(request),
        })

        task_name = getattr(request, 'task', None) if request else None
        task_args = getattr(request,
                            'argsrepr', getattr(request, 'args', None))
        task_kwargs = getattr(request,
                              'kwargsrepr', getattr(request, 'kwargs', None))
        worker = getattr(request, 'hostname', None)
        queue = request.delivery_info.get('routing_key')

        self.TaskModel._default_manager.store_result(
            content_type, content_encoding,
            task_id, result, status,
            traceback=traceback,
            meta=meta,
            task_name=task_name,
            task_args=task_args,
            task_kwargs=task_kwargs,
            worker=worker, queue=queue,
            using=using,
        )
        return result

    def _get_task_meta_for(self, task_id):
        """Get task metadata for a task by id."""
        obj = self.TaskModel._default_manager.get_task(task_id)
        res = obj.as_dict()
        meta = self.decode_content(obj, res.pop('meta', None)) or {}
        res.update(meta,
                   result=self.decode_content(obj, res.get('result')))
        return self.meta_from_decoded(res)

    def encode_content(self, data):
        content_type, content_encoding, content = self._encode(data)
        if content_encoding == 'binary':
            content = b64encode(content)
        return content_type, content_encoding, content

    def decode_content(self, obj, content):
        if content:
            if obj.content_encoding == 'binary':
                content = b64decode(content)
            return self.decode(content)

    def _forget(self, task_id):
        try:
            self.TaskModel._default_manager.get(task_id=task_id).delete()
        except self.TaskModel.DoesNotExist:
            pass

    def cleanup(self):
        """Delete expired metadata."""
        self.TaskModel._default_manager.delete_expired(self.expires)

    def add_to_chord(self, chord_id, result):
        raise NotImplementedError('Backend does not support add_to_chord')



from django.http import JsonResponse

from django_opstasks.common.datetime import current_datetime


class OpstasksResponse(JsonResponse):
    """
    return django.http.JsonResponse, setdefault safe=False

      error status code:
      - 400 Bad Request
      - 401 Unauthorized
      - 403 Forbidden
      - 404 Not Found
      - 405 Method Not Allowed

    """
    def __init__(self, context, status_code=200, safe=False, **kwargs):
        super().__init__(data=context, safe=safe, **kwargs)
        self.status_code = status_code


class AsyncResultResponse(OpstasksResponse):
    def __init__(self, async_result):
        data = {
            "timestamp": current_datetime(utc=False),
            "id": async_result.id,
            "name": async_result.name,
            "status": async_result.state,
            "args": async_result.args,
            "kwargs": async_result.kwargs,
            "queue": async_result.queue,
            "worker": async_result.worker,
        }
        super().__init__(data)



# Generated by Django 3.0.8 on 2020-09-01 11:34

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django_celery_beat.validators
import django_opstasks.models
import timezone_field.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ClockedSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('clocked_time', models.DateTimeField(help_text='在指定时间运行任务.', verbose_name='Clock Time')),
                ('enabled', models.BooleanField(default=True, editable=False, help_text='设置为False后不进行任务调度.', verbose_name='ENABLED')),
            ],
            options={
                'verbose_name': 'clocked',
                'verbose_name_plural': 'clocked',
                'db_table': 'tasks_clocked_schedule',
                'ordering': ['clocked_time'],
            },
        ),
        migrations.CreateModel(
            name='CrontabSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('minute', models.CharField(default='*', max_length=240, validators=[django_celery_beat.validators.minute_validator], verbose_name='Minutes')),
                ('hour', models.CharField(default='*', max_length=96, validators=[django_celery_beat.validators.hour_validator], verbose_name='Hours')),
                ('day_of_week', models.CharField(default='*', max_length=64, validators=[django_celery_beat.validators.day_of_week_validator], verbose_name='Days Of The Week')),
                ('day_of_month', models.CharField(default='*', max_length=124, validators=[django_celery_beat.validators.day_of_month_validator], verbose_name='Days Of The Month')),
                ('month_of_year', models.CharField(default='*', max_length=64, validators=[django_celery_beat.validators.month_of_year_validator], verbose_name='Months Of The Year')),
                ('timezone', timezone_field.fields.TimeZoneField(default='UTC', verbose_name='Timezone')),
            ],
            options={
                'verbose_name': 'crontab',
                'verbose_name_plural': 'crontabs',
                'db_table': 'tasks_crontab_schedule',
                'ordering': ['month_of_year', 'day_of_month', 'day_of_week', 'hour', 'minute', 'timezone'],
            },
        ),
        migrations.CreateModel(
            name='IntervalSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('every', models.IntegerField(help_text='再次运行任务之前要等待的间隔时间', validators=[django.core.validators.MinValueValidator(1)], verbose_name='间隔时间')),
                ('period', models.CharField(choices=[('days', '天'), ('hours', '小时'), ('minutes', '分钟'), ('seconds', '秒'), ('microseconds', '毫秒')], max_length=24, verbose_name='时间单位')),
            ],
            options={
                'verbose_name': 'interval',
                'verbose_name_plural': 'intervals',
                'db_table': 'tasks_interval_schedule',
                'ordering': ['period', 'every'],
                'unique_together': {('period', 'every')},
            },
        ),
        migrations.CreateModel(
            name='PeriodicTasks',
            fields=[
                ('ident', models.SmallIntegerField(default=1, primary_key=True, serialize=False, unique=True)),
                ('last_update', models.DateTimeField()),
            ],
            options={
                'db_table': 'tasks_periodic_tasks',
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('name', models.CharField(help_text='Name of the task.', max_length=50, primary_key=True, serialize=False, verbose_name='NAME')),
                ('Request', models.CharField(default='celery.worker.request:Request', help_text='Request class used, or the qualified name of one.', max_length=50, null=True, verbose_name='REQUEST')),
                ('Strategy', models.CharField(default='celery.worker.strategy:default', help_text='Execution strategy used, or the qualified name of one.', max_length=50, null=True, verbose_name='STRATEGY')),
                ('abstract', models.BooleanField(default=True, help_text='Deprecated attribute abstract here for compatibility.', null=True, verbose_name='ABSTRACT')),
                ('acks_late', models.BooleanField(default=False, help_text='When enabled messages for this task will be acknowledged after the task has been executed, and not just before (the default behavior).', null=True, verbose_name='ACKS LATE')),
                ('acks_on_failure_or_timeout', models.BooleanField(default=True, help_text='When enabled messages for this task will be acknowledged even if it fails or times out.', null=True, verbose_name='ACKS FAILURE')),
                ('app_main', models.CharField(default='opstasks', help_text='Celery application.', max_length=20, null=True, verbose_name='APP NAME')),
                ('autoregister', models.BooleanField(default=True, help_text='If disabled this task won’t be registered automatically.', null=True, verbose_name='AUTO REGISTER')),
                ('backend_url', models.CharField(default='mysql://user:passwd@localhost/opstasks', help_text='The result store backend class, or the name of the backend class to use.', max_length=100, null=True, verbose_name='BACKEND URL')),
                ('default_retry_delay', models.SmallIntegerField(default=180, help_text='Default time in seconds before a retry of the task should be executed. 3 minutes by default.', null=True, verbose_name='DEFAULT RETRY DELAY')),
                ('ignore_result', models.BooleanField(default=False, help_text='If enabled the worker won’t store task state and return values for this task.', null=True, verbose_name='IGNORE RESULT')),
                ('max_retries', models.SmallIntegerField(default=3, help_text='Maximum number of retries before giving up. If set to None, it will never stop retrying.', null=True, verbose_name='MAX RETRIES')),
                ('priority', models.SmallIntegerField(help_text='The task priority, a number between 0 and 9. .', null=True, verbose_name='PRIORITY')),
                ('rate_limit', models.CharField(help_text='None (no rate limit), "100/s" (hundred tasks a second),"100/m" (hundred tasks a minute), "100/h" (hundred tasks an hour)', max_length=10, null=True, verbose_name='RATE LIMIT')),
                ('reject_on_worker_lost', models.BooleanField(help_text='Even if acks_late is enabled, the worker will acknowledge tasks when the worker process executing them abruptly exits or is signaled (e.g., KILL/INT, etc).\nSetting this to true allows the message to be re-queued instead, so that the task will execute again by the same worker, or another worker.\nWarning: Enabling this can cause message loops; make sure you know what you’re doing.', null=True, verbose_name='REJECT ON LOST')),
                ('resultrepr_maxsize', models.SmallIntegerField(default=1024, help_text='Max length of result representation used in logs and events.', null=True, verbose_name='RESULTREPR MAXSIZE')),
                ('send_events', models.BooleanField(default=False, help_text='If enabled the worker will send monitoring events related to this task\nbut only if the worker is configured to send task related events', null=True, verbose_name='SEND EVENTS')),
                ('serializer', models.CharField(default='json', help_text='The name of a serializer that are registered with kombu.serialization.registry', max_length=10, null=True, verbose_name='SERIALIZER')),
                ('soft_time_limit', models.SmallIntegerField(help_text='Soft time limit. Defaults to the task_soft_time_limit setting.', null=True, verbose_name='SOFT TIME LIMIT')),
                ('store_errors_even_if_ignored', models.BooleanField(default=False, help_text='When enabled errors will be stored even if the task is otherwise configured to ignore results.', null=True, verbose_name='STORE ERRORS')),
                ('throws', models.CharField(default='()', help_text='Tuple of expected exceptions.', max_length=100, null=True, verbose_name='THROWS')),
                ('time_limit', models.SmallIntegerField(help_text='Hard time limit. Defaults to the task_time_limit setting.', null=True, verbose_name='TIME LIMIT')),
                ('track_started', models.BooleanField(default=False, help_text='If enabled the task will report its status as ‘started’ when the task is executed by a worker.', null=True, verbose_name='TRACK STARTED')),
                ('trail', models.BooleanField(default=True, help_text='If enabled the request will keep track of subtasks started by this task, and this information will be sent with the result (result.children).', null=True, verbose_name='TRAIL')),
                ('typing', models.BooleanField(default=True, help_text='Enable argument checking. You can set this to false if you don’t want the signature to be checked when calling the task. Defaults to Celery.strict_typing.', null=True, verbose_name='TYPING')),
                ('description', models.CharField(max_length=255, null=True, verbose_name='DESCRIPTION')),
                ('timestamp', models.DateTimeField(auto_now=True, help_text='Last update time of the task', verbose_name='TIMESTAMP')),
                ('status', models.IntegerField(choices=[(0, 'NORMAL'), (1, 'INVALID')], default=0, verbose_name='STATUS')),
            ],
            options={
                'verbose_name': '任务列表',
                'verbose_name_plural': '任务列表',
                'db_table': 'tasks_task',
            },
        ),
        migrations.CreateModel(
            name='TaskResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.CharField(db_index=True, help_text='Celery ID for the Task that was run', max_length=255, unique=True, verbose_name='Task ID')),
                ('task_name', models.CharField(db_index=True, help_text='Name of the Task which was run', max_length=255, null=True, verbose_name='Task Name')),
                ('task_args', models.TextField(help_text='JSON representation of the positional arguments used with the task', null=True, verbose_name='Task Positional Arguments')),
                ('task_kwargs', models.TextField(help_text='JSON representation of the named arguments used with the task', null=True, verbose_name='Task Named Arguments')),
                ('status', models.CharField(choices=[('FAILURE', 'FAILURE'), ('PENDING', 'PENDING'), ('RECEIVED', 'RECEIVED'), ('RETRY', 'RETRY'), ('REVOKED', 'REVOKED'), ('STARTED', 'STARTED'), ('SUCCESS', 'SUCCESS')], db_index=True, default='PENDING', help_text='Current state of the task being run', max_length=50, verbose_name='Task State')),
                ('worker', models.CharField(db_index=True, default=None, help_text='Worker that executes the task', max_length=100, null=True, verbose_name='Worker')),
                ('content_type', models.CharField(help_text='Content type of the result data', max_length=128, verbose_name='Result Content Type')),
                ('content_encoding', models.CharField(help_text='The encoding used to save the task result data', max_length=64, verbose_name='Result Encoding')),
                ('result', models.TextField(default=None, editable=False, help_text='The data returned by the task. Use content_encoding and content_type fields to read.', null=True, verbose_name='Result Data')),
                ('date_created', models.DateTimeField(auto_now_add=True, db_index=True, help_text='Datetime field when the task result was created in UTC', verbose_name='Created DateTime')),
                ('date_done', models.DateTimeField(auto_now=True, db_index=True, help_text='Datetime field when the task was completed in UTC', verbose_name='Completed DateTime')),
                ('traceback', models.TextField(blank=True, help_text='Text of the traceback if the task generated one', null=True, verbose_name='Traceback')),
                ('meta', models.TextField(default=None, editable=False, help_text='JSON meta information about the task, such as information on child tasks', null=True, verbose_name='Task Meta Information')),
            ],
            options={
                'verbose_name': '执行记录',
                'verbose_name_plural': '执行记录',
                'db_table': 'tasks_results',
            },
        ),
        migrations.CreateModel(
            name='SolarSchedule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('event', models.CharField(choices=[('dawn_astronomical', 'dawn_astronomical'), ('dawn_civil', 'dawn_civil'), ('dawn_nautical', 'dawn_nautical'), ('dusk_astronomical', 'dusk_astronomical'), ('dusk_civil', 'dusk_civil'), ('dusk_nautical', 'dusk_nautical'), ('solar_noon', 'solar_noon'), ('sunrise', 'sunrise'), ('sunset', 'sunset')], help_text='The type of solar event when the job should run', max_length=24, verbose_name='Solar Event')),
                ('latitude', models.DecimalField(decimal_places=6, help_text='Run the task when the event happens at this latitude', max_digits=9, validators=[django.core.validators.MinValueValidator(-90), django.core.validators.MaxValueValidator(90)], verbose_name='纬度')),
                ('longitude', models.DecimalField(decimal_places=6, help_text='Run the task when the event happens at this longitude', max_digits=9, validators=[django.core.validators.MinValueValidator(-180), django.core.validators.MaxValueValidator(180)], verbose_name='经度')),
            ],
            options={
                'verbose_name': 'solar',
                'verbose_name_plural': 'solars',
                'db_table': 'tasks_solar_schedule',
                'ordering': ('event', 'latitude', 'longitude'),
                'unique_together': {('event', 'latitude', 'longitude')},
            },
        ),
        migrations.CreateModel(
            name='PeriodicTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, unique=True, verbose_name='SHORT DESCRIPTION NAME')),
                ('task', models.CharField(max_length=200, verbose_name='TASK')),
                ('args', models.TextField(blank=True, help_text='传递给任务的位置参数, Json格式', validators=[django_opstasks.models.validate_tasks_args], verbose_name='位置参数')),
                ('kwargs', models.TextField(blank=True, help_text='传递给任务的关键字参数, Json格式', validators=[django_opstasks.models.validate_tasks_kwargs], verbose_name='任务参数')),
                ('queue', models.CharField(blank=True, default=None, help_text='选择CELERY_TASK_QUEUES中定义的队列, 留空将使用默认队列.', max_length=200, null=True, verbose_name='队列')),
                ('exchange', models.CharField(blank=True, default=None, help_text='Override Exchange for low-level AMQP routing', max_length=200, null=True, verbose_name='Exchange')),
                ('routing_key', models.CharField(blank=True, default=None, help_text='Override Routing Key for low-level AMQP routing', max_length=200, null=True, verbose_name='Routing Key')),
                ('headers', models.TextField(blank=True, default='{}', help_text='JSON encoded message headers for the AMQP message.', verbose_name='AMQP Message Headers')),
                ('priority', models.PositiveIntegerField(blank=True, default=None, help_text='优先级数字介于0-255之间. 仅支持: RabbitMQ, Redis . 优先级相反, 0最高.', null=True, validators=[django.core.validators.MaxValueValidator(255)], verbose_name='优先级')),
                ('start_time', models.DateTimeField(blank=True, help_text='计划的生效时间', null=True, verbose_name='开始时间')),
                ('expires', models.DateTimeField(blank=True, help_text='超过该时间之后, 本计划将不再启动任务', null=True, verbose_name='过期时间')),
                ('expire_seconds', models.PositiveIntegerField(blank=True, help_text='Timedelta with seconds which the schedule will no longer trigger the task to run', null=True, verbose_name='Expires timedelta with seconds')),
                ('enabled', models.BooleanField(default=True, help_text='取消勾选将关闭计划任务', verbose_name='ENABLED')),
                ('one_off', models.BooleanField(default=False, help_text='勾选后该计划仅执行一次', verbose_name='ONE_OFF')),
                ('last_run_at', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='Last Run Datetime')),
                ('total_run_count', models.PositiveIntegerField(default=0, editable=False, help_text='Running count of how many times the schedule has triggered the task', verbose_name='Total Run Count')),
                ('date_changed', models.DateTimeField(auto_now=True, help_text='Datetime that this PeriodicTask was last modified', verbose_name='Last Modified')),
                ('description', models.TextField(blank=True, default='Detailed description about the details of this Periodic Task', verbose_name='DESCRIPTION')),
                ('clocked', models.ForeignKey(blank=True, help_text='注:仅能设置一种调度方式,请将其余计划留空.', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_opstasks.ClockedSchedule', verbose_name='Clocked计划')),
                ('crontab', models.ForeignKey(blank=True, help_text='注:仅能设置一种调度方式,请将其余计划留空.', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_opstasks.CrontabSchedule', verbose_name='Crontab计划')),
                ('interval', models.ForeignKey(blank=True, help_text='注:仅能设置一种调度方式,请将其余计划留空.', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_opstasks.IntervalSchedule', verbose_name='Interval计划')),
                ('registered_task', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='TASK', to='django_opstasks.Task', verbose_name='已注册任务')),
                ('solar', models.ForeignKey(blank=True, help_text='注:仅能设置一种调度方式,请将其余计划留空.', null=True, on_delete=django.db.models.deletion.CASCADE, to='django_opstasks.SolarSchedule', verbose_name='Solar计划')),
            ],
            options={
                'verbose_name': '定时任务',
                'verbose_name_plural': '定时任务',
                'db_table': 'tasks_periodic_task',
            },
        ),
    ]
